@@ -107,6 +107,13 @@ void champlain_view_ensure_layers_visible (ChamplainView *view,
 
 void champlain_view_set_map_source (ChamplainView *view,
     ChamplainMapSource *map_source);
+void champlain_view_add_overlay_source (ChamplainView *view,
+    ChamplainMapSource *map_source,
+    guint8 opacity);
+void champlain_view_remove_overlay_source (ChamplainView *view,
+    ChamplainMapSource *map_source);
+GList *champlain_view_get_overlay_sources (ChamplainView *view);
+
 void champlain_view_set_deceleration (ChamplainView *view,
     gdouble rate);
 void champlain_view_set_kinetic_mode (ChamplainView *view,
